import bpy
import time
import os.path

class SimpleAudioVisualizerProperties(bpy.types.PropertyGroup):
    columns: bpy.props.IntProperty(
        name="Objects",
        default=20,
        min=1,
        description="Number of columns; higher values will increase calculation time"
    )
    path: bpy.props.StringProperty(
        name="Audio File Path",
        subtype='FILE_PATH',
        description="Path to the audio file to visualize"
    )
    mirror_mode: bpy.props.BoolProperty(
        name="Mirror Mode",
        default=False,
        description="Offsets the origins of the cubes to create a mirror effect"
    )


class SimpleAudioVisualizerOperator(bpy.types.Operator):
    bl_idname = "object.simple_audio_visualizer"
    bl_label = "Simple Audio Visualizer"
    bl_description = "Creates a simple audio visualizer"
    bl_options = {'REGISTER', 'UNDO'}

    def execute(self, context):
        sav_props = context.scene.sav_props

        start_time = time.time()

        columns = sav_props.columns
        total_objects = columns
        spectrum_Start = 10    # Frequency in Hz
        spectrum_End = 21000  # Frequency in Hz
        path = os.path.abspath(bpy.path.abspath(sav_props.path))

        mirror_mode = sav_props.mirror_mode

        maxheight = 6.0  # Default height value

        # For iterative frequency ranges
        l = 1
        h = spectrum_Start
        base = (spectrum_End / spectrum_Start) ** (1 / columns)

        print(f"STARTED (Cubes_Spectrum={columns})")

        bpy.ops.screen.frame_jump(end=False)

        # ---------------------------------------------------------------------
        # Create Visual Elements
        # ---------------------------------------------------------------------
        def create_visual_element(x, y, zmax, lo, hi):
            bpy.ops.mesh.primitive_cube_add(location=(x, y, 0))
            obj = bpy.context.active_object

            if mirror_mode:
                # Move origin to center of mass for the mirror effect
                bpy.ops.object.origin_set(type='ORIGIN_CENTER_OF_MASS', center='BOUNDS')
                obj.location.z += obj.dimensions.z / 2
            else:
                # Default origin at base of the cube
                bpy.context.scene.cursor.location = obj.location
                bpy.context.scene.cursor.location.z -= 1
                bpy.ops.object.origin_set(type='ORIGIN_CURSOR')

            # Scale and apply transform
            obj.scale.x = 0.4
            obj.scale.y = 0.4
            obj.scale.z = zmax
            bpy.ops.object.transform_apply(location=False, rotation=False, scale=True)

            # Insert a keyframe only for the Z-scale
            obj.keyframe_insert(data_path='scale', index=2)

            # Remove any X/Y scale curves
            action = obj.animation_data.action
            fcurves = action.fcurves
            for fc in fcurves:
                if fc.data_path == 'scale' and fc.array_index in [0, 1]:
                    fcurves.remove(fc)

            for fc in fcurves:
                fc.lock = False

            # Switch area to Graph Editor for the sound-to-fcurve operator
            bpy.context.area.type = 'GRAPH_EDITOR'
            fc_z_scale = fcurves.find('scale', index=2)
            fc_z_scale.select = True
            bpy.ops.graph.sound_to_samples(filepath=path, low=lo, high=hi)
            fc_z_scale.select = False

        # ---------------------------------------------------------------------
        # Generate the Cubes Along the Spectrum
        # ---------------------------------------------------------------------
        for i in range(columns):
            l = h
            h = round(spectrum_Start * base ** (i + 1), 2)
            print(f"     c: {i}   l: {l} Hz    h: {h} Hz")

            height = maxheight
            create_visual_element(0, i, height, l, h)

            # Optional progress display
            done_percent = (i + 1) / columns * 100
            elapsed_minutes = (time.time() - start_time) / 60
            print(f"{done_percent:.1f}%   done in {elapsed_minutes:.2f} minutes")

        print(f"CREATED {total_objects} OBJECTS")
        return {'FINISHED'}


class SimpleAudioVisualizerPanel(bpy.types.Panel):
    bl_label = "Simple Audio Visualizer"
    bl_idname = "GRAPH_EDITOR_PT_simple_audio_visualizer"
    bl_space_type = 'GRAPH_EDITOR'
    bl_region_type = 'UI'
    bl_category = "Simple Audio Visualizer"

    def draw(self, context):
        layout = self.layout
        sav_props = context.scene.sav_props

        layout.prop(sav_props, "columns")
        layout.prop(sav_props, "path")
        layout.prop(sav_props, "mirror_mode")

        layout.operator("object.simple_audio_visualizer", text="Create Visualizer")


# --------------------------------------------------------------------------
# Registration
# --------------------------------------------------------------------------
def register():
    bpy.utils.register_class(SimpleAudioVisualizerProperties)
    bpy.types.Scene.sav_props = bpy.props.PointerProperty(type=SimpleAudioVisualizerProperties)

    bpy.utils.register_class(SimpleAudioVisualizerOperator)
    bpy.utils.register_class(SimpleAudioVisualizerPanel)


def unregister():
    bpy.utils.unregister_class(SimpleAudioVisualizerPanel)
    bpy.utils.unregister_class(SimpleAudioVisualizerOperator)

    del bpy.types.Scene.sav_props
    bpy.utils.unregister_class(SimpleAudioVisualizerProperties)
